' Some general constants
%TRUE   = 1
%FALSE  = 0
%MAX_TOKEN_WIDTH = 256 ' no keyword, identifier name, string etc.. can be longer than this

' Special characters and their ASCII value
%SPACE      = 32
%CR         = 13
%LF         = 10
%UNDERSCORE = 95
%DOT        = 46
%SMALLA     = 97
%SMALLZ     = 122
%BIGA       = 65
%BIGZ       = 90
%ZERO       = 48
%NINE       = 57
%TAB        = 9
%COMMENT    = 39  ' single quote
%EOF        = 0   ' end of file marker
%EQUALS     = 61  ' equals sign
%GT         = 62  ' greater than
%LT         = 60  ' less than
%ASTERISK   = 42
%SLASH      = 47  ' /
%BACKSLASH  = 92  ' \
%PLUS       = 43
%MINUS      = 45
%DOUBLEQUOTE= 34

' Types of tokens
%TOKEN_IDENT      = 1

%TOKEN_INTEGER    = 2
%TOKEN_FLOATING   = 3
%TOKEN_COMMENT    = 4
%TOKEN_STRING    = 5

%TOKEN_FOR        = 201
%TOKEN_TO         = 202
%TOKEN_NEXT       = 203
%TOKEN_WHILE      = 204
%TOKEN_WEND       = 205

%TOKEN_PLUS       = 101
%TOKEN_MINUS      = 101
%TOKEN_TIMES      = 103
%TOKEN_DIV        = 104
%TOKEN_INTDIV     = 105
%TOKEN_EQUALS     = 106
%TOKEN_NOTEQUALS  = 107
%TOKEN_GT         = 108
%TOKEN_LT         = 109
%TOKEN_GTEQ       = 110
%TOKEN_LTEQ       = 111

%TOKEN_EOF        = 665
%TOKEN_ERROR      = 666

' Definition of a token
Type Token_Type
  TokType As Dword
  TokValue As String * %MAX_TOKEN_WIDTH
End Type

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Global Variables
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Global LookAhead As String    ' The look ahead character
Global FileNum As Long ' File Handle (created by Startup(), destroyed by shutdown
Global CurrentLine As Dword ' The current line being processed
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

' Return the text reporentation of a token
Function GetTokenName(ByVal TokType As Dword) As String
  Dim TheName As String

  Select Case TokType
  Case %TOKEN_IDENT
    Function = "[IDENTIFIER]"
  Case %TOKEN_FOR
    Function = "[FOR       ]"
  Case %TOKEN_TO
    Function = "[TO        ]"
  Case %TOKEN_NEXT
    Function = "[NEXT      ]"
  Case %TOKEN_WHILE
    Function = "[WHILE     ]"
  Case %TOKEN_WEND
    Function = "[WEND      ]"
  Case %TOKEN_INTEGER
    Function = "[INTEGER   ]"
  Case %TOKEN_FLOATING
    Function = "[FLOATING  ]"
  Case %TOKEN_COMMENT
    Function = "[COMMENT   ]"
  Case %TOKEN_PLUS
    Function = "[PLUS      ]"
  Case %TOKEN_MINUS
    Function = "[MINUS     ]"
  Case %TOKEN_TIMES
    Function = "[TIMES     ]"
  Case %TOKEN_DIV
    Function = "[DIV       ]"
  Case %TOKEN_INTDIV
    Function = "[INTDIV    ]"
  Case %TOKEN_EQUALS
    Function = "[EQUALS    ]"
  Case %TOKEN_NOTEQUALS
    Function = "[NOTEQUALS ]"
  Case %TOKEN_GT
    Function = "[GREATER   ]"
  Case %TOKEN_LT
    Function = "[LESS      ]"
  Case %TOKEN_GTEQ
    Function = "[GREATER-EQ]"
  Case %TOKEN_LTEQ
    Function = "[LESS_EQ   ]"
  Case %TOKEN_EOF
    Function = "[EOF       ]"
  Case %TOKEN_STRING
    Function = "[STRING    ]"
  Case %TOKEN_ERROR
    Function = "[ERROR     ]"
  Case Else
    Function = "[UNKNOWN!!!]"
  End Select

End Function

' print an error
Sub PrintError(ByVal Text As String)
  Print "ERROR: " & Text & " (Line:" & Trim$(Str$(CurrentLine)) & ")"
End Sub

' print a warining
Sub PrintWarning(ByVal Text As String)
  Print "WARINING: " & Text & " (Line:" & Trim$(Str$(CurrentLine)) & ")"
End Sub

' Determine whether a character is alpha (a letter)
Function IsAlpha(ByVal ch As String) As Dword
  Dim AscCh As Dword
  AscCh = Asc(ch)     ' ASCII value of the char

  If (AscCh >= %SMALLA And AscCh <= %SMALLZ) _
      Or _
     (AscCh >= %BIGA And AscCh <= %BIGZ) Then

     Function = %TRUE
  Else
     Function = %FALSE
  End If
End Function

' Determine whether a character is a digit
Function IsDigit(ByVal ch As String) As Dword
  Dim AscCh As Dword
  AscCh = Asc(ch)     ' ASCII value of the char

  If (AscCh >= %ZERO And AscCh <= %NINE) Then
     Function = %TRUE
  Else
     Function = %FALSE
  End If
End Function

' Determine whether a character is whitespace
Function IsWhiteSpace(ByVal ch As String) As Dword
  Dim AscCh As Dword
  AscCh = Asc(ch)     ' ASCII value of the char

  If (AscCh = %SPACE Or AscCh = %TAB Or AscCh = %LF Or AscCh = %CR) Then
     Function = %TRUE
  Else
     Function = %FALSE
  End If
End Function

' Determine whether a character is valid in an identifier
Function IsIdentChar(ByVal ch As String) As Dword
   If IsAlpha(ch) = %TRUE Or IsDigit(ch) = %TRUE Or ch = Chr$(%UNDERSCORE) Then
      Function = %TRUE
   Else
      Function = %FALSE
   End If
End Function

' determine whether a character is valid in a number
Function IsNumChar(ByVal ch As String) As Dword
   If IsDigit(ch) = %TRUE Or ch = Chr$(%DOT) Then
      Function = %TRUE
   Else
      Function = %FALSE
   End If
End Function

' determine whether a character is valid in an operator
Function IsOpChar(ByVal ch As String) As Dword
   If ch = Chr$(%EQUALS) Or _
      ch = Chr$(%GT) Or _
      ch = Chr$(%LT) Or _
      ch = Chr$(%ASTERISK) Or _
      ch = Chr$(%SLASH) Or _
      ch = Chr$(%BACKSLASH) Or _
      ch = Chr$(%PLUS) Or _
      ch = Chr$(%MINUS) Then

      Function = %TRUE
   Else
      Function = %FALSE
   End If
End Function

' get a character from the input stream
Function GetChar() As String

  ' read a character from the input stream
  Dim ch As String
  Get$ FileNum,1,ch

  Select Case Asc(ch)
    Case %CR ' we hit the end of a line

      Incr CurrentLine ' mark the next line
      Function = ch

    Case -1 ' Get$ returns -1 if we passed the end of file

      Function = Chr$(%EOF)

    Case Else
      Function = ch ' the next char

  End Select
End Function

' get the next significant character from the input (one
' that is not a white space)
Function GetNextSigChar() As String
  Dim ch As String

  Do
    ch = GetChar()
  Loop Until IsWhiteSpace(ch) = %FALSE

  Function = ch

End Function

' determine token value of a string
' if 'Text' is a Keyword then return the token corresponding
' to that keyword, otherwise it must be an identifier
Function GetTokenValue(ByVal Text As String) As Dword

  Select Case UCase$(Text)
    Case "FOR"
      Function = %TOKEN_FOR
    Case "TO"
      Function = %TOKEN_TO
    Case "NEXT"
      Function = %TOKEN_NEXT
    Case "WHILE"
      Function = %TOKEN_WHILE
    Case "WEND"
      Function = %TOKEN_WEND
    Case Else
      Function = %TOKEN_IDENT
  End Select

End Function

' read a string
' ch is the starting character
Sub GetString(ByRef Tok As Token_Type, ByVal ch As String)
  Dim StringToken As String

  StringToken = ""

  If Asc(ch) <> %DOUBLEQUOTE Then
         PrintError "Missing opening quote."
         Tok.TokType = %TOKEN_ERROR
         Tok.TokValue = ""

         ' read the next lookahead just in case
         LookAhead = GetChar()
         Exit Sub
  Else
    'read the next char
    ch = GetChar()
  End If

  ' read up to the closing quote
  While Asc(ch) <> %DOUBLEQUOTE

    ' if we hit the end of line we're in trouble because there is no closing quote
    If Asc(ch) = %LF Then
         PrintError "Missing closing quote."
         Tok.TokType = %TOKEN_ERROR
         Tok.TokValue = ""

         ' read the next lookahead just in case
         LookAhead = GetChar()
         Exit Sub
    End If

    StringToken = StringToken + ch
    ch = GetChar() ' read the next char
  Wend

  If Len(StringToken) > %MAX_TOKEN_WIDTH Then
     StringToken = Left$(StringToken,%MAX_TOKEN_WIDTH)
     PrintWarning "String too long - truncating."
  End If

  ' in this case we didn't read the extra lookahead
  ' because the last char we read was the closing quote
  ' which is not valid, so read another one
  LookAhead = GetChar()

  ' return the token
  Tok.TokType = %TOKEN_STRING

  Tok.TokValue = StringToken
End Sub

' read an identifier or keyword
' ch is the starting character
Sub GetWord(ByRef Tok As Token_Type, ByVal ch As String)
  Dim WordToken As String

  WordToken = ""

  While IsIdentChar(ch)
    WordToken = WordToken + ch
    ch = GetChar() ' read the next char
  Wend

  ' remember we read an extra char
  LookAhead = ch

  ' return the token
  Tok.TokType = GetTokenValue(WordToken)

  Tok.TokValue = WordToken
End Sub

' read a number
' ch is the starting character
Sub GetNumber(ByRef Tok As Token_Type, ByVal ch As String)
  Dim WordToken As String, DotFound As Dword

  WordToken = ""
  DotFound = %FALSE

  While IsNumChar(ch) = %TRUE

    ' take note if we find a .
    ' but if we find more then one we're in trouble
    If Asc(ch) = %DOT Then
      If DotFound = %TRUE Then
         PrintError "Invaild floating point value."
         Tok.TokType = %TOKEN_ERROR
         Tok.TokValue = ""

         ' read the next lookahead just in case
         LookAhead = GetChar()
         Exit Sub
      Else
        DotFound = %TRUE
      End If
    End If

    WordToken = WordToken + ch
    ch = GetChar() ' read the next char
  Wend

  ' if the last character is a . then the number is invalid
  ' we cannot have a floating point number like 12.
  If Right$(WordToken,1) = "." Then
         PrintError "Invaild floating point value."
         Tok.TokType = %TOKEN_ERROR
         Tok.TokValue = ""

         ' read the next lookahead just in case
         LookAhead = GetChar()
         Exit Sub
  End If

  ' remember we read an extra char
  LookAhead = ch

  ' return the token
  If DotFound = %TRUE Then
    Tok.TokType = %TOKEN_FLOATING
  Else
    Tok.TokType = %TOKEN_INTEGER
  End If

  Tok.TokValue = WordToken
End Sub

' read (and basically ignore) a comment till the end of the line
' ch is the starting character
Sub GetComment(ByRef Tok As Token_Type, ByVal ch As String)
  Dim WordToken As String

  WordToken = ""

  ' Munch up the comment
  While Asc(ch) <> %LF
    ch = GetChar() ' read the next char
  Wend

  ' remember we read an extra char
  LookAhead = ch

  Tok.TokType = %TOKEN_COMMENT
  Tok.TokValue = ""
End Sub

' read an operator
' ch is the starting character
Sub GetOperator(ByRef Tok As Token_Type, ByVal ch As String)
  ' valid operators are
  '    +   -   *   /   \   =   >   <   >=   <=   <>

  Select Case ch
    Case "+"
      Tok.TokType = %TOKEN_PLUS
      Tok.TokValue = ""

      ' remember to read the extra lookahead
      LookAhead = GetChar()

    Case "-"
      Tok.TokType = %TOKEN_MINUS
      Tok.TokValue = ""

      ' remember to read the extra lookahead
      LookAhead = GetChar()

    Case "*"
      Tok.TokType = %TOKEN_TIMES
      Tok.TokValue = ""

      ' remember to read the extra lookahead
      LookAhead = GetChar()

    Case "/"
      Tok.TokType = %TOKEN_DIV
      Tok.TokValue = ""

      ' remember to read the extra lookahead
      LookAhead = GetChar()

    Case "\"
      Tok.TokType = %TOKEN_INTDIV
      Tok.TokValue = ""

      ' remember to read the extra lookahead
      LookAhead = GetChar()

    Case "="
      Tok.TokType = %TOKEN_EQUALS
      Tok.TokValue = ""

      ' remember to read the extra lookahead
      LookAhead = GetChar()

    Case ">"  ' operator may have more than one char e.g. >=
      ' read an extra char to see what's next
      ch = GetChar()

      If ch = "=" Then
        ' its a Greater-Or-Equal-To !!
        Tok.TokType = %TOKEN_GTEQ
        Tok.TokValue = ""

        ' remember to read the extra lookahead
        LookAhead = GetChar()

      Else
        ' its only a Greater-Than !!
        Tok.TokType = %TOKEN_GT
        Tok.TokValue = ""

        ' we already read the extra char uselessly
        LookAhead = ch

      End If

    Case "<"  ' operator may have more than one char e.g. <=, <>
      ' read an extra char to see what's next
      ch = GetChar()

      If ch = "=" Then
        ' its a Less-Or-Equal-To !!
        Tok.TokType = %TOKEN_LTEQ
        Tok.TokValue = ""

        ' remember to read the extra lookahead
        LookAhead = GetChar()

      ElseIf ch = ">" Then
        ' its a Not-Equal-To !!
        Tok.TokType = %TOKEN_NOTEQUALS
        Tok.TokValue = ""

        ' remember to read the extra lookahead
        LookAhead = GetChar()

      Else
        ' its only a Less-Than !!
        Tok.TokType = %TOKEN_LT
        Tok.TokValue = ""

        ' we already read the extra char uselessly
        LookAhead = ch

      End If

  End Select
End Sub

' read a token from the input
Sub GetNextToken(ByRef Tok As Token_Type)

  Dim ch As String
  ch = LookAhead

  ' get the next significant character
  If IsWhiteSpace(ch) Then ch = GetNextSigChar()

  If Asc(ch) = %EOF Then
    ' we're done scanning
    Tok.TokType = %TOKEN_EOF
    Tok.TokValue = ""

  ElseIf IsAlpha(ch) Then
    ' we're dealing with an identifier or keyword
    Call GetWord(Tok, ch)

  ElseIf IsNumChar(ch) Then
    ' we're dealing with a number
    Call GetNumber(Tok, ch)

  ElseIf IsOpChar(ch) Then
    ' we're dealing with an operator
    Call GetOperator(Tok, ch)

  ElseIf Asc(ch) = %DOUBLEQUOTE Then
    ' we're dealing with a string
    Call GetString(Tok, ch)

  ElseIf Asc(ch) = %COMMENT Then
    ' we're dealing with a comment
    Call GetComment(Tok, ch)
  Else
    ' we're dealing with an illegal character
    PrintError "Unknown character '" & ch & "' in source file."

    ' return an error token
    Tok.TokType = %TOKEN_ERROR
    Tok.TokValue = ""
  End If

End Sub

' Initialise the lexical analyser with the required source file
Function StartUp(SourceFile As String) As Dword
  On Error GoTo ErrorHandler

  ' set the current line and character
  CurrentLine = 1 ' we start off from line 1

  ' open the source file
    ' check if the source file exists
    If Len(Dir$(SourceFile)) = 0 Then
      PrintError "Source file not found."
      GoTo ErrorHandler
    End If

    FileNum = FreeFile
    Open SourceFile For Binary As FileNum

  ' read the first character
  LookAhead = GetChar

  Function = %TRUE
  Exit Function
ErrorHandler:
  If Err Then
    PrintError "Error initialising lexical analyser."
  End If

  Function = %FALSE
End Function

' clean up
Sub ShutDown()
  ' close the source file
  Close FileNum
End Sub

Function PbMain()

  Print "Lexical Analyser - Kristian Guillaumier 2001"
  Print "--------------------------------------------"
  Print

  Dim Tok As Token_Type

  ' Initialise everything

  ' read the input file from the command line
  Dim filename As String
  filename = Command$

  If StartUp(filename) = %TRUE Then

    ' If everything initalised well proceed

    Do
      Call GetNextToken(Tok)

      ? GetTokenName(Tok.TokType) & " " & Trim$(Tok.TokValue)

    Loop Until Tok.TokType = %TOKEN_EOF Or Tok.TokType = %TOKEN_ERROR

  End If

  ' Clean up
  Call ShutDown()

  WaitKey$
End Function



